use crate::tmpfs::TempDir;
use errors::{BinError, BuildError, InstallError};
use log::{debug, trace};
use manifest::Manifest;
use pkgfile::PKGFile;
use std::fmt::Display;

pub mod errors;

#[derive(Debug)]
pub enum InstallType {
    Build,
    Bin,
}

#[derive(Debug)]
pub struct PackageInstaller {
    manifest: Manifest,
    pkgfile: PKGFile,
    install_type: InstallType,
}

impl PackageInstaller {
    pub fn new(m: Manifest, p: PKGFile, i: InstallType) -> PackageInstaller {
        PackageInstaller {
            manifest: m,
            pkgfile: p,
            install_type: i,
        }
    }

    fn extract_to(&self, path: String) -> Result<(), BinError> {
        if !self.pkgfile.has_data() {
            return Err(BinError::UnpackError("package has no data".to_string()));
        }
        if std::path::Path::new(&path).exists() {
            return Err(BinError::UnpackError("path already exists".to_string()));
        }
        tar::Archive::new(self.pkgfile.data.as_slice())
            .unpack(&path)
            .map_err(|e| BinError::UnpackError(e.to_string()))
    }

    fn bin(&self) -> Result<(), BinError> {
        let mut tmpdir = TempDir::default();
        tmpdir.push(&self.manifest.package.name);
        trace!("extracting package into: {}", tmpdir.to_string());
        match self.extract_to(tmpdir.to_string()) {
            Ok(_) => {}
            Err(e) => return Err(e),
        }
        debug!("extracted package in: {}", tmpdir.to_string());
        Ok(())
    }

    fn build(&self) -> Result<(), BuildError> {
        if let None = self.manifest.build.clone() {
            return Err(BuildError::InvalidManifest);
        }
        let build_manifest = self.manifest.build.clone().unwrap();
        // TODO: Check dependencies

        Ok(())
    }

    pub fn install(&self) -> Result<(), InstallError> {
        match self.install_type {
            InstallType::Bin => self.bin().map_err(|e| InstallError::BinError(e)),
            InstallType::Build => self.build().map_err(|e| InstallError::BuildError(e)),
        }
    }
}
